# Import libraries.
import os
import numpy as np
import time
import math
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Activation, Dropout
from tensorflow.keras import optimizers, losses, activations
from tensorflow.keras.callbacks import CSVLogger

# Import source files.
from descriptor_sets import *

###############################################################################
# Load data.
###############################################################################

def load_data(mass_transport_property, number_of_samples_train, number_of_samples_val, number_of_samples_test):
    
    # Folder.
    folder_data = "../data"
    
    # Load data.
    x_train = np.fromfile(os.path.join(folder_data, "descriptors_train.bin"), dtype = np.float32)
    x_val = np.fromfile(os.path.join(folder_data, "descriptors_val.bin"), dtype = np.float32)
    x_test = np.fromfile(os.path.join(folder_data, "descriptors_test.bin"), dtype = np.float32)

    y_train = np.fromfile(os.path.join(folder_data, mass_transport_property + "_train.bin"), dtype = np.float32)
    y_val = np.fromfile(os.path.join(folder_data, mass_transport_property + "_val.bin"), dtype = np.float32)
    y_test = np.fromfile(os.path.join(folder_data, mass_transport_property + "_test.bin"), dtype = np.float32)

    # Reshape data.
    number_of_descriptors = x_train.size // number_of_samples_train

    x_train = np.reshape(x_train, (number_of_samples_train, number_of_descriptors))
    x_val = np.reshape(x_val, (number_of_samples_val, number_of_descriptors))
    x_test = np.reshape(x_test, (number_of_samples_test, number_of_descriptors))

    y_train = np.reshape(y_train, (number_of_samples_train, 1))
    y_val = np.reshape(y_val, (number_of_samples_val, 1))
    y_test = np.reshape(y_test, (number_of_samples_test, 1))

    # Return data sets.
    return x_train, x_val, x_test, y_train, y_val, y_test

###############################################################################
# Preprocess data.
###############################################################################

def preprocess_data(mass_transport_property, x_train, x_val, x_test, y_train, y_val, y_test):

    number_of_descriptors = x_train.shape[1]

    # Normalize the input data to zero mean and unit variance.
    for current_descriptor in range(0, number_of_descriptors):
        m_x = np.mean(x_train[:, current_descriptor])
        s_x = np.std(x_train[:, current_descriptor])

        x_train[:, current_descriptor] = (x_train[:, current_descriptor] - m_x) / s_x
        x_val[:, current_descriptor] = (x_val[:, current_descriptor] - m_x) / s_x
        x_test[:, current_descriptor] = (x_test[:, current_descriptor] - m_x) / s_x

    # Transform output data. Logit for m, log for k.
    if mass_transport_property == "m":
        y_train = np.log(y_train / (1 - y_train))
        y_val = np.log(y_val / (1 - y_val))
        y_test = np.log(y_test / (1 - y_test))
    else:
        y_train = np.log(y_train)
        y_val = np.log(y_val)
        y_test = np.log(y_test)

    # Return data sets.
    return x_train, x_val, x_test, y_train, y_val, y_test

###############################################################################
# Define ANN model.
###############################################################################

def ann_model(number_of_descriptors, number_of_layers, number_of_nodes_per_layer, activation):

    model = Sequential()
    model.add(Dense(number_of_nodes_per_layer, activation = activation, input_shape = (number_of_descriptors,)))
    for counter in range(0, number_of_layers - 1):
        model.add(Dense(number_of_nodes_per_layer, activation = activation))
    model.add(Dense(1, activation = 'linear'))
    model.summary()

    return model

###############################################################################
# Main function.
###############################################################################

if __name__ == '__main__':
    
    # Pick mass transport property.
    mass_transport_property = "m"
    # mass_transport_property = "kappa"

    # Folders.
    folder_trained_models = "trained_models_" + mass_transport_property
    folder_predictions = "predictions_" + mass_transport_property

    # Data set sizes.
    number_of_samples_train = 63000
    number_of_samples_val = 13500
    number_of_samples_test = 13500

    # Load data.
    x_train_full, x_val_full, x_test_full, y_train, y_val, y_test = load_data(mass_transport_property, number_of_samples_train, number_of_samples_val, number_of_samples_test)
    
    # Preprocess data.
    x_train_full, x_val_full, x_test_full, y_train, y_val, y_test = preprocess_data(mass_transport_property, x_train_full, x_val_full, x_test_full, y_train, y_val, y_test)

    # Loop over descriptor sets.
    descriptor_sets = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])

    for descriptor_set in descriptor_sets:

        # Extract descriptor set.
        idx_descriptors = create_descriptor_idx_vector(descriptor_set)   

        x_train = x_train_full[:, idx_descriptors]
        x_val = x_val_full[:, idx_descriptors]
        x_test = x_test_full[:, idx_descriptors]
        number_of_descriptors = x_train.shape[1]
    
        # Model parameters.
        number_of_layers = 4
        number_of_nodes_per_layer = 64
        activation = 'elu'

        # Define model.
        model = ann_model(number_of_descriptors, number_of_layers, number_of_nodes_per_layer, activation)

        # Load model weights.
        file_path_model = folder_trained_models + "/" + "model_" + str(descriptor_set) + ".hdf5"
        model.load_weights(file_path_model)
        
        model.summary()

        # Predict.
        batch_size = 2048
        y_train_hat = model.predict(x_train, batch_size = batch_size, verbose = 1)
        y_val_hat = model.predict(x_val, batch_size = batch_size, verbose = 1)
        y_test_hat = model.predict(x_test, batch_size = batch_size, verbose = 1)
    
        # Save results.
        y_train_hat = np.asarray(y_train_hat, dtype = np.float32)
        y_val_hat = np.asarray(y_val_hat, dtype = np.float32)
        y_test_hat = np.asarray(y_test_hat, dtype = np.float32)

        y_train_hat.tofile(folder_predictions + "/" "y_train_hat_" + str(descriptor_set) + ".bin")
        y_val_hat.tofile(folder_predictions + "/" "y_val_hat_" + str(descriptor_set) + ".bin")
        y_test_hat.tofile(folder_predictions + "/" "y_test_hat_" + str(descriptor_set) + ".bin")
