# Import libraries.
import os
import numpy as np
import time
import math
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Activation, Dropout
from tensorflow.keras import optimizers, losses, activations
from tensorflow.keras.callbacks import CSVLogger

# Import source files.
from descriptor_sets import *

###############################################################################
# Load data.
###############################################################################

def load_data(mass_transport_property, number_of_samples_train, number_of_samples_val):
    
    # Folder.
    folder_data = "../data"
    
    # Load data.
    x_train = np.fromfile(os.path.join(folder_data, "descriptors_train.bin"), dtype = np.float32)
    x_val = np.fromfile(os.path.join(folder_data, "descriptors_val.bin"), dtype = np.float32)

    y_train = np.fromfile(os.path.join(folder_data, mass_transport_property + "_train.bin"), dtype = np.float32)
    y_val = np.fromfile(os.path.join(folder_data, mass_transport_property + "_val.bin"), dtype = np.float32)

    # Reshape data.
    number_of_descriptors = x_train.size // number_of_samples_train

    x_train = np.reshape(x_train, (number_of_samples_train, number_of_descriptors))
    x_val = np.reshape(x_val, (number_of_samples_val, number_of_descriptors))

    y_train = np.reshape(y_train, (number_of_samples_train, 1))
    y_val = np.reshape(y_val, (number_of_samples_val, 1))

    # Return data sets.
    return x_train, x_val, y_train, y_val

###############################################################################
# Preprocess data.
###############################################################################

def preprocess_data(mass_transport_property, x_train, x_val, y_train, y_val):

    number_of_descriptors = x_train.shape[1]

    # Normalize the input data to zero mean and unit variance.
    for current_descriptor in range(0, number_of_descriptors):
        m_x = np.mean(x_train[:, current_descriptor])
        s_x = np.std(x_train[:, current_descriptor])

        x_train[:, current_descriptor] = (x_train[:, current_descriptor] - m_x) / s_x
        x_val[:, current_descriptor] = (x_val[:, current_descriptor] - m_x) / s_x

    # Transform output data. Logit for m, log for k.
    if mass_transport_property == "m":
        y_train = np.log(y_train / (1 - y_train))
        y_val = np.log(y_val / (1 - y_val))
    else:
        y_train = np.log(y_train)
        y_val = np.log(y_val)

    # Return data sets.
    return x_train, x_val, y_train, y_val

###############################################################################
# Define ANN model.
###############################################################################

def ann_model(number_of_descriptors, number_of_layers, number_of_nodes_per_layer, activation):

    model = Sequential()
    model.add(Dense(number_of_nodes_per_layer, activation = activation, input_shape = (number_of_descriptors,)))
    for counter in range(0, number_of_layers - 1):
        model.add(Dense(number_of_nodes_per_layer, activation = activation))
    model.add(Dense(1, activation = 'linear'))
    model.summary()

    return model

###############################################################################
# Learning rate schedule.
###############################################################################

def learning_rate_schedule(current_epoch, log10_learning_rate_vector, number_of_epochs_vector):
   
    log10_learning_rate = log10_learning_rate_vector[0]
    for current_step in range(1, number_of_epochs_vector.size):
        if current_epoch >= np.cumsum(number_of_epochs_vector)[current_step-1]:
            log10_learning_rate = log10_learning_rate_vector[current_step]

    return 10 ** log10_learning_rate
    
###############################################################################
# Main function.
###############################################################################

if __name__ == '__main__':
    
    # Start time.
    start_time = time.time()

    # Random seed.
    random_seed = np.mod(1e7 * start_time, np.iinfo(np.uint32).max)
    random_seed = np.uint32(random_seed)
    np.random.seed(random_seed)

    # Pick mass transport property.
    mass_transport_property = "m"
    #mass_transport_property = "kappa"

    # Data set sizes.
    number_of_samples_train = 63000
    number_of_samples_val = 13500

    # Load data.
    x_train_full, x_val_full, y_train, y_val = load_data(mass_transport_property, number_of_samples_train, number_of_samples_val)
    
    # Preprocess data.
    x_train_full, x_val_full, y_train, y_val = preprocess_data(mass_transport_property, x_train_full, x_val_full, y_train, y_val)

    # Pick random descriptor set.
    descriptor_set = np.random.choice([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    idx_descriptors = create_descriptor_idx_vector(descriptor_set)   

    x_train = x_train_full[:, idx_descriptors]
    x_val = x_val_full[:, idx_descriptors]
    number_of_descriptors = x_train.shape[1]
    
    # File paths for storing results.
    folder_results = os.path.join("training_results_" + mass_transport_property, "descriptor_set_" + str(descriptor_set))
    if not os.path.exists(folder_results):
        os.mkdir(folder_results)
        
    file_path_model = os.path.join(folder_results, "training_" + str(random_seed) + "_model.hdf5")
    file_path_losses = os.path.join(folder_results, "training_" + str(random_seed) + "_losses.csv")
    file_path_parameters = os.path.join(folder_results, "training_" + str(random_seed) + "_parameters.dat")

    # Model parameters.
    number_of_layers = 4
    number_of_nodes_per_layer = 64
    activation = 'elu'

    # Training parameters.
    batch_size = 128
    validation_batch_size = 2048
    momentum = 0.9
    nesterov = False

    log10_learning_rate_vector = np.array([-4.50, -4.00, -3.50, -3.00, -2.50, -2.00, -2.50, -3.00, -3.50, -4.00])
    number_of_epochs_vector    = np.array([ 1000,  1000,  1000,  1000,  1000,  4000,  4000,  4000,  4000,  4000]) 
    
    number_of_epochs = np.sum(number_of_epochs_vector)

    # Save model and training parameters.
    text_file = open(file_path_parameters, "w")
    text_file.write("Number of layers = " + str(number_of_layers) + "\n")
    text_file.write("Number of nodes per layer = " + str(number_of_nodes_per_layer) + "\n")
    text_file.write("Activation = " + activation + "\n")
    text_file.write("Batch size = " + str(batch_size) + "\n")
    text_file.write("Momentum = " + str(momentum) + "\n")
    text_file.write("Nesterov = " + str(nesterov) + "\n")
    text_file.write("log10 learning rate vector = " + str(log10_learning_rate_vector) + "\n")
    text_file.write("Number of epochs vector = " + str(number_of_epochs_vector) + "\n")
    text_file.close()

    # Define model.
    model = ann_model(number_of_descriptors, number_of_layers, number_of_nodes_per_layer, activation)

    # Callbacks.
    callback_learning_rate = keras.callbacks.LearningRateScheduler(lambda current_epoch: learning_rate_schedule(current_epoch, log10_learning_rate_vector, number_of_epochs_vector)) 
    callback_model_checkpoint = keras.callbacks.ModelCheckpoint(file_path_model, monitor = 'val_loss', verbose = 0, save_best_only = True, save_weights_only = True, mode = 'auto', save_freq = 'epoch')
    callback_csv_logger = CSVLogger(file_path_losses, append = True, separator = ';')

    # Training.
    sgd = optimizers.SGD(momentum = momentum, nesterov = nesterov)
    model.compile(loss = losses.mean_squared_error, optimizer = sgd)

    model.fit(  x = x_train, 
                y = y_train,
                validation_data = (x_val, y_val),
                epochs = number_of_epochs,
                batch_size = batch_size,
                validation_batch_size = validation_batch_size,
                callbacks = [callback_learning_rate, callback_model_checkpoint, callback_csv_logger],
                verbose = 2)

    # Elapsed time.
    elapsed_time = time.time() - start_time

    # Append execution time to parameter file.
    text_file = open(file_path_parameters, "a")
    text_file.write("Execution time (s) = " + str(np.round(elapsed_time, decimals = 2)) + "\n")
    text_file.close()
